//---------------------------------------------------------------------------

#include <vcl.h>
#include <math.h>
#include <math.hpp>

#pragma hdrstop

#include "Geometry.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

typedef double T3D[4];

typedef double TMatrix[4][4];

void Zero3D(T3D& d) {
  d[0]=d[1]=d[2]=0;
  d[3]=1;
};

void Set3D(T3D& d, double x, double y, double z, double w) {
  d[0]=x; d[1]=y; d[2]=z; d[3]=w;
};


void Copy3D(T3D& d, T3D s) {
  for (int i=0; i<4; i++)
    d[i]=s[i];
};

void Add3D(T3D& d, T3D s) {
  for (int i=0; i<3; i++)
    d[i]+=s[i];
};

void Mpy3D(T3D& d, double s) {
  for (int i=0; i<3; i++)
    d[i]*=s;
};

void Neg3D(T3D& d) {
  for (int i=0; i<3; i++)
    d[i]=-d[i];
};

void Identity(TMatrix& m) {
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      if (i==j) m[i][j]=1;
      else m[i][j]=0;
};

void CopyMatrix(TMatrix& d, TMatrix s) {
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      d[i][j]=s[i][j];
};

void MpyMatrix(TMatrix& d, TMatrix s) {
  double sum;
  TMatrix tmp;
  CopyMatrix(tmp,d);
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++) {
      sum=0;
      for (int k=0; k<4; k++)
        sum+=tmp[i][k]*s[k][j];
      d[i][j]=sum;
    }
};

void TransMatrix(TMatrix& d) {
  TMatrix tmp;
  CopyMatrix(tmp,d);
  for (int i=0; i<4; i++)
    for (int j=0; j<4; j++)
      d[i][j]=tmp[j][i];
};

void Translate(TMatrix& m, T3D s) {
  TMatrix tmp;
  Identity(tmp);
  for (int i=0; i<3; i++)
    tmp[3][i]=s[i];
  MpyMatrix(m,tmp);
};

void RotateX(TMatrix& m, double angle) {
  TMatrix tmp;
  Identity(tmp);
  Extended s,c;
  SinCos(DegToRad(angle),s,c);
  tmp[1][1]=tmp[2][2]=(double)c;
  tmp[1][2]=(double)s; tmp[2][1]=-(double)s;
  MpyMatrix(m,tmp);
};

void RotateY(TMatrix& m, double angle) {
  TMatrix tmp;
  Identity(tmp);
  Extended s,c;
  SinCos(DegToRad(angle),s,c);
  tmp[0][0]=tmp[2][2]=(double)c;
  tmp[0][2]=-(double)s; tmp[2][0]=(double)s;
  MpyMatrix(m,tmp);
};

void RotateZ(TMatrix& m, double angle) {
  TMatrix tmp;
  Identity(tmp);
  Extended s,c;
  SinCos(DegToRad(angle),s,c);
  tmp[0][0]=tmp[1][1]=(double)c;
  tmp[0][1]=(double)s; tmp[1][0]=-(double)s;
  MpyMatrix(m,tmp);
};

void RotateDOrigin(TMatrix& m, double angle, T3D vd) {
  double d=sqrt(vd[0]*vd[0]+vd[2]*vd[2]);
  if (d==0) { RotateY(m,angle); return; }
  double D=sqrt(vd[0]*vd[0]+vd[1]*vd[1]+vd[2]*vd[2]);
  double dzd=vd[2]/d;
  double dxd=vd[0]/d;
  double dyD=vd[1]/D;
  double dD=d/D;
  TMatrix r1,r2,tmp;
  Identity(r1);
  Identity(r2);
  r1[0][0]=r1[2][2]=dzd;
  r1[0][2]=dxd; r1[2][0]=-dxd;
  r2[1][1]=r2[2][2]=dD;
  r2[1][2]=dyD; r2[2][1]=-dyD;
  Identity(tmp);
  MpyMatrix(tmp,r1);
  MpyMatrix(tmp,r2);
  RotateZ(tmp,angle);
  TransMatrix(r2);
  TransMatrix(r1);
  MpyMatrix(tmp,r2);
  MpyMatrix(tmp,r1);
  MpyMatrix(m,tmp);
};

void Transform3D(T3D& d, TMatrix t) {
  double sum;
  T3D tmp;
  Copy3D(tmp,d);
  for (int i=0; i<4; i++) {
    sum=0;
    for (int j=0; j<4; j++)
      sum+=tmp[j]*t[j][i];
    d[i]=sum;
  }
};

